from weightInitialization import initConv, initConvB, initW, initB, zeros
from activationFunctions import relu
from theano.sandbox.cuda.dnn import dnn_conv

import theano
import theano.tensor as T

import numpy as np
import itertools

from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams


from theano.sandbox.cuda.basic_ops import gpu_alloc_empty
from theano.sandbox.cuda.dnn import GpuDnnConvDesc, GpuDnnConvGradI

srng = RandomStreams()

def dropout(X, p):
    if p > 0:
        retain_prob = 1 - p
        X *= srng.binomial(X.shape, p=retain_prob, dtype=theano.config.floatX)
        X /= retain_prob
    return X

def corropt(X, p): 
    if p > 0:
        retain_prob = 1 - p
        X *= srng.binomial(X.shape, p=retain_prob, dtype=theano.config.floatX)
    return X

class CaeConvolutionEncoder: 
    def __init__(self, phi, filter_shape, stride, name='c', p_corropt=0., parameters=None):
        """
        :type filter_shape: tuple or list of length 4
        :param filter_shape: (number of filters, num input feature maps, filter height, filter width)
        """

        phi = corropt(phi, p_corropt)

        ## Network parameters
        if parameters: 
            W, b = parameters
        else: 
            W = initConv(name + 'W', filter_shape)
            b = initConvB(name + 'b', filter_shape)

        ## Perform calculations
        conv_phi = dnn_conv(
            img=phi,
            kerns=W,
            subsample=stride,
            border_mode='valid'
        )
        out = conv_phi + b.dimshuffle('x', 0, 'x', 'x')
        self.output = relu(out)

        ## Store parameters
        self.theta = [W, b]

class CaeConvolutionDecoder:
    def __init__(self, phi, filter_shape, stride, upsampleSize, name='c', border_mode='full'):
        """
        :type filter_shape: tuple or list of length 4
        :param filter_shape: (number of filters, num input feature maps, filter height, filter width)
        """

        ## Network parameters
        d0,d1,d2,d3 = filter_shape
        fs = (d1,d0,d2,d3)
        W = initConv(name + 'W', fs)
        b = initConvB(name + 'b', fs)

        if True: 
            upsample = T.zeros(upsampleSize)
            upsample = T.set_subtensor(upsample[:, :, ::stride[0], ::stride[1]], phi)

            ## Perform calculations
            conv_phi = dnn_conv(
                img=upsample,
                kerns=W,
                border_mode=border_mode)
            out = conv_phi + b.dimshuffle('x', 0, 'x', 'x')
            self.output = relu(out)
        else: 
            ## Code from https://github.com/Newmu/dcgan_code/blob/master/lib/ops.py#L85-L96
            img = phi
            kerns = W
            subsample = stride
            conv_mode = 'conv'
            border_mode = (0,0)

            desc = GpuDnnConvDesc(
                border_mode=border_mode, 
                subsample=subsample,
                conv_mode=conv_mode)(
                    gpu_alloc_empty(
                        img.shape[0], 
                        kerns.shape[1], 
                        img.shape[2]*subsample[0], 
                        img.shape[3]*subsample[1]).shape, kerns.shape)
            out = gpu_alloc_empty(
                img.shape[0], 
                kerns.shape[1], 
                img.shape[2]*subsample[0], 
                img.shape[3]*subsample[1])
            d_img = GpuDnnConvGradI()(kerns, img, out, desc)

            self.output = relu(d_img + b.dimshuffle('x', 0, 'x', 'x'))

        ## Store parameters
        self.theta = [W, b]

class CaeHiddenEncoder: 
    def __init__(self, phi, n_in, n_out, name='h', actFn=relu, p_corropt=0., parameters=None):

        phi = corropt(phi, p_corropt)

        ## Network parameters
        if parameters: 
            W, b = parameters
        else: 
            W = initW(name + 'W', n_in, n_out)
            b = initB(name + 'b', n_in, n_out)

        ## Perform calculations 
        self.output = actFn(T.dot(phi, W) + b)

        ## Store parameters
        self.theta = [W, b]
        #self.W = W


class InversLayer:
    def __init__(self, phi, phi_in, phi_out): 
        self.output = theano.grad(None, wrt=phi_in, known_grads={phi_out: phi})


class CaeOutputLayer: 
    def __init__(self, phi_org, phi_dec, grad_clip=1): 
        phi_dec = phi_dec[:, :, :-1, :-1] ## HACK: phi_dec.shape=(b,4,85,85) -> phi_dec.shape=(b,4,84,84)
        loss = T.sum((phi_dec-phi_org)**2, axis=(1,2,3))
        loss = T.mean(loss)
        # oss = theano.gradient.grad_clip(loss, -grad_clip, grad_clip)
        self._loss = loss
        self._phi_org = phi_org
        self._phi_dec = phi_dec

    def loss(self): 
        return self._loss

    def org_dec(self): 
        return self._phi_org, self._phi_dec



## TEST CODE

def testCae(): 
    inp1 = T.tensor4(dtype='float32')
    inp2 = T.tensor4(dtype='float32')
    x = np.random.randn(2, 4, 84, 84).astype(np.float32)

    filter_shape = (32, 4, 8, 8)

    print 'Compiling functions'
    enc_layer = CaeConvolutionEncoder(inp1, filter_shape, (4,4))
    dec_layer = InversLayer(inp2, inp1, enc_layer.output)

    enc = theano.function([inp1], enc_layer.output)
    dec = theano.function([inp1,inp2], dec_layer.output)

    print 'Encoding x '
    e = enc(x)
    print 'Decoding x'
    d = dec(x, e)
    print "Decoded shape: ", d.shape
    print "Differnce between org and dec:", np.sum(x-d)

def testHidd(): 
    inp1 = T.matrix(dtype='float32')
    inp2 = T.matrix(dtype='float32')
    x = np.random.randn(32, 128).astype(np.float32)

    print 'Compiling functions'
    enc_layer = CaeHiddenEncoder(inp1, n_in=128, n_out=50)
    dec_layer = InversLayer(inp2, inp1, enc_layer.output)

    enc = theano.function([inp1], enc_layer.output)
    dec = theano.function([inp1,inp2], dec_layer.output)

    print 'Encoding x '
    e = enc(x)
    print 'Decoding x'
    d = dec(x, e)
    print "Decoded shape: ", d.shape
    print "Differnce between org and dec:", np.sum(x-d)

def testTraining(): 
    inp1 = T.tensor4(dtype='float32')
    x = np.random.randn(2, 4, 84, 84).astype(np.float32)

    filter_shape = (32, 4, 8, 8)

    print 'Compiling functions'
    enc_layer = CaeConvolutionEncoder(inp1, filter_shape, (4,4))
    inp2 = enc_layer.output
    dec_layer = InversLayer(inp2, inp1, enc_layer.output)

    enc = theano.function([inp1], enc_layer.output)
    dec = theano.function([inp1], dec_layer.output)

    loss = T.mean((inp1-dec_layer.output)**2)
    theta = enc_layer.theta
    grads = theano.grad(loss, theta)
    updates = [(p, p-0.1*g) for p, g in zip(theta, grads)]

    trainFn = theano.function([inp1], loss, updates=updates)

    
    print "Differnce between org and dec:", np.sum(x-dec(x))

    for i in range(100): 
        print 'Loss:', trainFn(x)

    print "Differnce between org and dec:", np.sum(x-dec(x))

def testMultiLayered(): 
    inp1 = T.tensor4(dtype='float32')
    x = np.random.randn(2, 4, 84, 84).astype(np.float32)

    filter_shape1 = (32, 4, 8, 8)
    filter_shape2 = (64, 32, 4, 4)

    print 'Compiling functions'
    enc_l1 = CaeConvolutionEncoder(inp1, filter_shape1, (4,4))
    enc_l2 = CaeConvolutionEncoder(enc_l1.output, filter_shape2, (2,2))
    enc_out = enc_l2.output
    dec_layer = InversLayer(enc_out, inp1, enc_out)

    enc = theano.function([inp1], enc_out)
    dec = theano.function([inp1], dec_layer.output)

    print "Differnce between org and dec:", np.sum(x-dec(x))

def testAll(): 
    phi = T.tensor4(dtype='float32')
    # x = np.random.uniform(0,0.1,(2, 4, 84, 84)).astype(np.float32)

    x = np.zeros((4,4,84,84), dtype='float32')
    h = 42

    x[0,:,0:h,0:h] = 1
    x[1,:,0:h,h:] = 1
    x[2,:,h:,0:h] = 1
    x[3,:,h:,h:] = 1

    filter_shape1 = (32, 4, 8, 8)
    filter_shape2 = (64, 32, 4, 4)
    filter_shape3 = (64, 64, 3, 3)


    enc_l1 = CaeConvolutionEncoder(phi, filter_shape1, (4,4), name='cae_c1')
    enc_l2 = CaeConvolutionEncoder(enc_l1.output, filter_shape2, (2,2), name='cae_c2')
    enc_l3 = CaeConvolutionEncoder(enc_l2.output, filter_shape3, (1,1), name='cae_c3')
    enc_l4 = CaeHiddenEncoder(T.flatten(enc_l3.output, outdim=2), 3136, 512, name='cae_e_hid') 
    enc_out = enc_l4.output

    if False:
        ## -> Get the fucking nan problem
        dec_layer = InversLayer(enc_out, phi, enc_out)
        o_layer = CaeOutputLayer(phi, dec_layer.output)
        layers = [enc_l1, enc_l2, enc_l3, enc_l4]
    else: 
        dec_l4 = CaeHiddenEncoder(enc_out, 512, 3136, name='cae_d4_hid')
        dec_l5 = CaeHiddenEncoder(dec_l4.output, 3136, 8096, name='cae_d5_hid')
        dec_l6 = CaeHiddenEncoder(dec_l5.output, 8096, 28224, name='cae_d5_hid')
        dec_out = T.reshape(dec_l6.output, (4,4,84,84), ndim=4)
        o_layer = CaeOutputLayer(phi, dec_out)
        layers = [enc_l1, enc_l2, enc_l3, enc_l4, dec_l4, dec_l5, dec_l6]


    loss = o_layer.loss()
    theta = list( itertools.chain( *[layer.theta for layer in layers] ) )
    grads = theano.grad(loss, theta)

    updates = []
    for p, g in zip(theta, grads): 
        shape = p.get_value(borrow=True).shape

        acc_ms = zeros(name='acc_ms', shape=shape)
        acc_ms_new = 0.95 * acc_ms + 0.05 * g ** 2 

        delta = - (0.00025 * g)/(T.sqrt(acc_ms_new+0.00001))

        p_new = p + delta

        updates.append((acc_ms, acc_ms_new))
        updates.append((p, p_new))
    

    if False:
        theano.printing.pydotprint(dec_layer.output, outfile='pics/dec.png', var_with_name_simple=True)


    print 'Compiling functions'
    _, dec = o_layer.org_dec()
    dec = theano.function([phi], dec)

    trainFn = theano.function([phi], [loss, enc_out], updates=updates)

    print 'Decoding: ', dec(x).shape
    
    print "Differnce between org and dec:", np.mean(x-dec(x))

    for i in range(500):
        loss, enc = trainFn(x)
        print 'Epoch=%i, Loss=%f' % (i, loss)
        #e = np.array(enc)
        #print np.min(e, axis=1), np.max(e, axis=1)

    print "Differnce between org and dec:", np.mean(x-dec(x))




if __name__ == '__main__':
    testAll()
    # testTraining()