import theano.tensor as T
import numpy as np
from weightInitialization import initConv, initConvB
from weightInitialization import initW, initB, copyTensor
from activationFunctions import relu
from theano.sandbox.cuda.dnn import dnn_conv

class DqnInputLayer: 
    def __init__(self, phi, phi_n): 
        self.online_phi_out = phi
        self.online_phi_n_out = phi_n
        self.target_phi_n_out = phi_n


class DqnConvolutionalLayer:
    def __init__(self, preceding, filter_shape, stride=None, name='c', actFn=relu, border_mode='valid'):
        """
        :type filter_shape: tuple or list of length 4
        :param filter_shape: (#filters, #input feature maps, filter height, filter width)
        """

        online_phi   = preceding.online_phi_out
        online_phi_n = preceding.online_phi_n_out
        target_phi_n = preceding.target_phi_n_out

        ## Network parameters
        oW = initConv(name + 'oW', filter_shape)
        ob = initConvB(name + 'ob', filter_shape)

        ## Target network parameters
        tW = copyTensor(oW, name + 'tW', prefix=False)
        tb = copyTensor(ob, name + 'tb', prefix=False)

        ## Perform calculations
        conv_online_phi_out = dnn_conv(
            img=online_phi,
            kerns=oW,
            subsample=stride,
            border_mode=border_mode
        )

        conv_online_phi_n_out = dnn_conv(
            img=online_phi_n,
            kerns=oW,
            subsample=stride,
            border_mode=border_mode
        )

        conv_target_phi_n_out = dnn_conv(
            img=target_phi_n,
            kerns=tW,
            subsample=stride,
            border_mode=border_mode
        )

        self.online_phi_out   = actFn(conv_online_phi_out + ob.dimshuffle('x', 0, 'x', 'x'))
        self.online_phi_n_out = actFn(conv_online_phi_n_out + ob.dimshuffle('x', 0, 'x', 'x'))
        self.target_phi_n_out = actFn(conv_target_phi_n_out + tb.dimshuffle('x', 0, 'x', 'x'))

        ## Store parameters
        self.oTheta = [oW, ob]
        self.tTheta = [tW, tb]


class DqnFlattenLayer:
    def __init__(self, preceding, outdim=2): 
        online_phi   = preceding.online_phi_out
        online_phi_n = preceding.online_phi_n_out
        target_phi_n = preceding.target_phi_n_out

        self.online_phi_out = T.flatten(online_phi, outdim=outdim)
        self.online_phi_n_out = T.flatten(online_phi_n, outdim=outdim)
        self.target_phi_n_out = T.flatten(target_phi_n, outdim=outdim)


class DqnDenseLayer():
    def __init__(self, preceding, n_in, n_out, name='h', actFn=relu):
        online_phi   = preceding.online_phi_out
        online_phi_n = preceding.online_phi_n_out
        target_phi_n = preceding.target_phi_n_out

        ## Network parameters
        oW = initW(name + 'oW', n_in, n_out)
        ob = initB(name + 'ob', n_in, n_out)

        ## Target network parameters
        tW = copyTensor(oW, name + 'tW', prefix=False)
        tb = copyTensor(ob, name + 'tb', prefix=False)

        ## Perform calculations 
        self.online_phi_out = actFn(T.dot(online_phi, oW) + ob)
        self.online_phi_n_out = actFn(T.dot(online_phi_n, oW) + ob)
        self.target_phi_n_out = actFn(T.dot(target_phi_n, tW) + tb)

        ## Store parameters
        self.oTheta = [oW, ob]
        self.tTheta = [tW, tb]


class DqnOutputLayer:
    def __init__(self, preceding, n_in, n_out, batchSize, gamma, a, r, t, 
            name='o', clip_reward=1, clip_delta=1, batch_acc='mean', doubleDQN=True):
        online_phi   = preceding.online_phi_out
        online_phi_n = preceding.online_phi_n_out
        target_phi_n = preceding.target_phi_n_out

        ## Network parameters
        oW = initW(name + 'oW', n_in, n_out)
        ob = initB(name + 'ob', n_in, n_out)

        ## Target network parameters
        tW = copyTensor(oW, name + 'tW', prefix=False)
        tb = copyTensor(ob, name + 'tb', prefix=False)

        ## Perform calculations
        oav_phi = T.dot(online_phi, oW) + ob
        oav_phi_n = T.dot(online_phi_n, oW) + ob
        tav_phi_n = T.dot(target_phi_n, tW) + tb

        ## Store parameters
        self.oTheta = [oW, ob]
        self.tTheta = [tW, tb]

        ####################################
        ### Creating the action function ###
        #################################### 
        self._oav_phi_mean = T.mean(oav_phi)
        self._action = T.argmax(oav_phi, axis=1)[0]

        ##################################
        ### Creating the loss function ###
        ##################################
        offset = np.arange(0, n_out*batchSize, n_out)
        if clip_reward > 0: 
            r = T.clip(r, -clip_reward, clip_reward)

        if doubleDQN: 
            a_n = T.argmax(oav_phi_n, axis=1)
            y = r + (1-t) * gamma * T.take(tav_phi_n, a_n+offset)
        else: 
            y = r + (1-t) * gamma * T.max(tav_phi_n, axis=1) ## r + (1-t) * gamma * max Q(s', a)
        qX = T.take(oav_phi, a+offset) ## Q(s, a)
        delta = y - qX

        if clip_delta > 0: 
            delta = T.clip(delta, -clip_delta, clip_delta)
            
        square_delta = delta**2
        if batch_acc == 'sum':
            loss = T.sum(square_delta)
        elif batch_acc == 'mean': 
            loss = T.mean(square_delta)
        else: 
            raise ValueError( 'Unrecognized batch accumilator, batch_acc="%s". ' % str(batch_acc) ) 

        self._loss = loss

    def loss(self):
        return self._loss 

    def action(self):
        return self._action

    def actionValueMean(self): 
        return self._oav_phi_mean